from __future__ import annotations

import asyncio
import time
from collections import defaultdict
from typing import TYPE_CHECKING, Any, TypeAlias

import async_timeout
import disnake
from disnake.abc import Connectable
from disnake.utils import MISSING
from loguru import logger

import lavamystic
from lavamystic.utils import ExtrasNamespace
from lavamystic.exceptions import (
    ChannelTimeoutException,
    InvalidChannelStateException,
    LavalinkException,
    QueueEmpty,
)
from lavamystic.filters import Filters
from lavamystic.node import Pool
from lavamystic.payloads import (
    PlayerUpdateEventPayload,
    TrackEndEventPayload,
)
from lavamystic.queue import Queue
from lavamystic.tracks import Playable, Playlist

if TYPE_CHECKING:
    from disnake import TextChannel, Thread, VoiceChannel, StageChannel
    from disnake.types.voice import GuildVoiceState as GuildVoiceStatePayload
    from disnake.types.gateway import VoiceServerUpdateEvent
    from typing_extensions import Self

    from .node import Node
    from .types.request import Request as RequestPayload
    from .types.state import PlayerVoiceState, VoiceState

    VocalGuildChannel = disnake.VoiceChannel | disnake.StageChannel

    t_channel: TypeAlias = (
        TextChannel |
        Thread |
        VoiceChannel |
        StageChannel
    )


T_a: TypeAlias = list[Playable] | Playlist


class Player(disnake.VoiceProtocol):
    """The Player is a :class:`disnake.VoiceProtocol` used to connect your :class:`disnake.Client` to a
    :class:`disnake.VoiceChannel`.

    The player controls the music elements of the bot including playing tracks, the queue, connecting etc.
    See Also: The various methods available.

    .. note::

        Since the Player is a :class:`disnake.VoiceProtocol`, it is attached to the various ``voice_client`` attributes
        in disnake.py, including ``guild.voice_client``, ``ctx.voice_client`` and ``interaction.voice_client``.

    Attributes
    ----------
    queue: :class:`~Queue`
        The queue associated with this player.
    auto_queue: :class:`~Queue`
        The auto_queue associated with this player.
        This queue holds tracks that are recommended by the AutoPlay feature.
    """

    channel: VocalGuildChannel

    def __call__(self, client: disnake.Client, channel: VocalGuildChannel) -> Self:
        super().__init__(client, channel)

        self._guild = channel.guild

        return self

    def __init__(
        self, client: disnake.Client = MISSING, channel: Connectable = MISSING, *, nodes: list[Node] | None = None
    ) -> None:
        super().__init__(client, channel)

        self.client: disnake.Client = client
        self._guild: disnake.Guild | None = None

        self._voice_state: PlayerVoiceState = {"voice": {}}

        self._node: Node
        if not nodes:
            self._node = Pool.get_node()
        else:
            self._node = sorted(nodes, key=lambda n: len(n.players))[0]

        if self.client is MISSING and self.node.client:
            self.client = self.node.client

        self._last_update: int | None = None
        self._last_position: int = 0
        self._ping: int = -1

        self._connected: bool = False
        self._connection_event: asyncio.Event = asyncio.Event()

        self._current: Playable | None = None
        self._original: Playable | None = None
        self._previous: Playable | None = None

        self.queue: Queue = Queue()
        self.auto_queue: Queue = Queue()

        self._volume: int = 100
        self._paused: bool = False

        self._auto_cutoff: int = 20
        self._auto_weight: int = 3
        self._previous_seeds_cutoff: int = self._auto_cutoff * self._auto_weight
        self._history_count: int | None = None

        self._autoplay: lavamystic.AutoPlayMode = lavamystic.AutoPlayMode.partial
        self.__previous_seeds: asyncio.Queue[str] = asyncio.Queue(maxsize=self._previous_seeds_cutoff)

        self._auto_lock: asyncio.Lock = asyncio.Lock()
        self._error_count: int = 0

        self._filters: Filters = Filters()

        self._inactivity_task: asyncio.Task[bool] | None = None
        self._inactivity_wait: int | None = self._node.inactive_timeout

        self._paused_task: asyncio.Task[bool] | None = None
        self._paused_wait: int | None = self._node.resume_timeout

        self._namespace: dict[str, any] = defaultdict()

    def dispatch_message_update(self) -> None:
        if not self._current:
            return

        self.client.dispatch("mystic_message_update", self)

    @property
    def namespace(self) -> ExtrasNamespace:
        return ExtrasNamespace(self._namespace)

    def add_to_namespace(self, value: dict[str, any], /) -> None:
        if not isinstance(value, dict):
            raise TypeError(f"Expected a dict, got {type(value)}")

        for key, _value in dict(value).items():
            if self._namespace.get(key) and _value is None:
                del self._namespace[key]
                continue

            self._namespace[key] = _value

    def _inactivity_task_callback(self, task: asyncio.Task[bool]) -> None:
        assert self._guild
        result: bool = task.result()
        cancelled: bool = task.cancelled()

        if cancelled or result is False:
            logger.debug(f"Disregarding Check Task {task.get_name()} as it was previously cancelled.")
            return

        if not result:
            logger.debug(f"Disregarding Check Task {task.get_name()} as it received an unknown result.")
            return

        if self.playing:
            logger.debug(
                f"Disregarding Check Task {task.get_name()} as Player {self._guild.id} is playing.",
            )
            return

        self.client.dispatch("mystic_inactive_player", self)
        logger.debug(f'Dispatched "on_mystic_inactive_player" for Player {self._guild.id}.')

    @staticmethod
    async def _runner(wait: int) -> bool:
        try:
            await asyncio.sleep(wait)
        except asyncio.CancelledError:
            return False
        return True

    def inactivity_cancel(self) -> None:
        if self._inactivity_task:
            try:
                self._inactivity_task.cancel()
            except Exception as e:
                _ = e
                pass

        self._inactivity_task = None

    def inactivity_start(self) -> None:
        if self._inactivity_wait is not None and self._inactivity_wait > 0:
            self._inactivity_task = asyncio.create_task(self._runner(self._inactivity_wait))
            self._inactivity_task.add_done_callback(self._inactivity_task_callback)

    def track_start(self) -> None:
        self.inactivity_cancel()

    async def auto_play_event(self, payload: TrackEndEventPayload) -> None:
        if self._autoplay is lavamystic.AutoPlayMode.disabled:
            self.inactivity_start()
            return

        if self._error_count >= 3:
            logger.warning(
                "AutoPlay was unable to continue as you have received too many consecutive errors."
                "Please check the error log on Lavalink."
            )
            self.inactivity_start()
            return

        if payload.reason == "replaced":
            self._error_count = 0
            return

        elif payload.reason == "loadFailed":
            self._error_count += 1

        else:
            self._error_count = 0

        if self.node.status is not lavamystic.NodeStatus.CONNECTED:
            logger.warning(f'"Unable to use AutoPlay on Player for Guild "{self.guild}" due to disconnected Node.')
            return

        if not isinstance(self.queue, Queue) or not isinstance(self.auto_queue, Queue):  # type: ignore
            logger.warning(f'"Unable to use AutoPlay on Player for Guild "{self.guild}" due to unsupported Queue.')
            self.inactivity_start()
            return

        if self.queue.mode is lavamystic.QueueMode.loop:
            await self._do_partial(history=False)

        elif self.queue.mode is lavamystic.QueueMode.loop_all:
            await self._do_partial()

        elif self._autoplay is lavamystic.AutoPlayMode.partial or self.queue:
            await self._do_partial()

    async def _do_partial(self, *, history: bool = True) -> None:
        self.inactivity_start()

        if self._current is None:
            try:
                track: Playable = self.queue.get()
            except QueueEmpty:
                return

            await self.play(track, add_history=history)

    # TODO: Recommendations maybe later

    @property
    def inactive_timeout(self) -> int | None:
        """A property which returns the time as an ``int`` of seconds to wait before this player dispatches the
        :func:`on_mystic_inactive_player` event.

        This property could return ``None`` if no time has been set.

        An inactive player is a player that has not been playing anything for the specified amount of seconds.

        - Pausing the player while a song is playing will not activate this countdown.
        - The countdown starts when a track ends and cancels when a track starts.
        - The countdown will not trigger until a song is played for the first time or this property is reset.
        - The default countdown for all players is set on :class:`~Node`.

        This property can be set with a valid ``int`` of seconds to wait before dispatching the
        :func:`on_mystic_inactive_player` event or ``None`` to remove the timeout.


        .. warning::
            Setting this to a value of ``0`` or below is the equivalent of setting this property to ``None``.


        When this property is set, the timeout will reset, and all previously waiting countdowns are cancelled.

        - See: :class:`~Node`
        - See: :func:`on_mystic_inactive_player`
        """
        return self._inactivity_wait

    @inactive_timeout.setter
    def inactive_timeout(self, value: int | None) -> None:
        if not value or value <= 0:
            self._inactivity_wait = None
            self.inactivity_cancel()
            return

        if value < 10:
            logger.warning('Setting "inactive_timeout" below 10 seconds may result in unwanted side effects.')

        self._inactivity_wait = value
        self.inactivity_cancel()

        if self.connected and not self.playing:
            self.inactivity_start()

    @property
    def autoplay(self) -> lavamystic.AutoPlayMode:
        """A property which returns the :class:`AutoPlayMode` the player is currently in.

        This property can be set with any :class:`AutoPlayMode` enum value.


        .. versionchanged:: 3.0.0

            This property now accepts and returns a :class:`AutoPlayMode` enum value.
        """
        return self._autoplay

    @autoplay.setter
    def autoplay(self, value: Any) -> None:
        if not isinstance(value, lavamystic.AutoPlayMode):
            raise ValueError("Please provide a valid 'AutoPlayMode' to set.")

        self._autoplay = value

    @property
    def node(self) -> Node:
        """The :class:`Player`'s currently selected :class:`Node`.
        """
        return self._node

    @property
    def guild(self) -> disnake.Guild | None:
        """Returns the :class:`Player`'s associated :class:`disnake.Guild`.

        Could be None if this :class:`Player` has not been connected.
        """
        return self._guild

    @property
    def connected(self) -> bool:
        """Returns a bool indicating if the player is currently connected to a voice channel.
        """
        return self.channel and self._connected

    @property
    def current(self) -> Playable | None:
        """Returns the currently playing :class:`~Playable` or None if no track is playing."""
        return self._current

    @property
    def volume(self) -> int:
        """Returns an int representing the currently set volume, as a percentage.

        See: :meth:`set_volume` for setting the volume.
        """
        return self._volume

    @property
    def filters(self) -> Filters:
        """Property which returns the :class:`~Filters` currently assigned to the Player.

        See: :meth:`~Player.set_filters` for setting the players filters.

        .. versionchanged:: 3.0.0

            This property was previously known as ``filter``.
        """
        return self._filters

    @property
    def paused(self) -> bool:
        """Returns the paused status of the player. A currently paused player will return ``True``.

        See: :meth:`pause` and :meth:`play` for setting the paused status.
        """
        return self._paused

    @property
    def ping(self) -> int:
        """Returns the ping in milliseconds as int between your connected Lavalink Node and disnake (Players Channel).

        Returns ``-1`` if no player update event has been received or the player is not connected.
        """
        return self._ping

    @property
    def playing(self) -> bool:
        """Returns whether the :class:`~Player` is currently playing a track and is connected.

        Due to relying on validation from Lavalink, this property may in some cases return ``True`` directly after
        skipping/stopping a track, although this is not the case when disconnecting the player.

        This property will return ``True`` in cases where the player is paused *and* has a track loaded.
        """
        return self._connected and self._current is not None and not self._paused

    @property
    def position(self) -> int:
        """Returns the position of the currently playing :class:`~Playable` in milliseconds.

        This property relies on information updates from Lavalink.

        In cases there is no :class:`~Playable` loaded or the player is not connected,
        this property will return ``0``.

        This property will return ``0`` if no update has been received from Lavalink.
        """
        if self.current is None or not self.playing:
            return 0

        if not self.connected:
            return 0

        if self._last_update is None:
            return 0

        if self.paused:
            return self._last_position

        position: int = int((time.monotonic_ns() - self._last_update) / 1000000) + self._last_position
        return min(position, self.current.length)

    async def update_event(self, payload: PlayerUpdateEventPayload) -> None:
        self._last_update = time.monotonic_ns()
        self._last_position = payload.position

        self._ping = payload.ping

    async def on_voice_state_update(self, data: GuildVoiceStatePayload, /) -> None:
        channel_id = data["channel_id"]

        if not channel_id:
            return await self.destroy()

        self._connected = True

        self._voice_state["voice"]["session_id"] = data["session_id"]
        self.channel = self.client.get_channel(int(channel_id))  # type: ignore

    async def on_voice_server_update(self, data: VoiceServerUpdateEvent, /) -> None:
        self._voice_state["voice"]["token"] = data["token"]
        self._voice_state["voice"]["endpoint"] = data["endpoint"]

        await self._dispatch_voice_update()

    async def _dispatch_voice_update(self) -> None:
        assert self.guild is not None
        data: VoiceState = self._voice_state["voice"]

        try:
            session_id: str = data["session_id"]
            token: str = data["token"]
        except KeyError:
            return

        endpoint: str | None = data.get("endpoint", None)
        if not endpoint:
            return

        request: RequestPayload = {"voice": {"sessionId": session_id, "token": token, "endpoint": endpoint}}

        try:
            await self.node.update(self.guild.id, data=request)
        except LavalinkException:
            await self.disconnect()
        else:
            self._connection_event.set()

        logger.debug(f"Player {self.guild.id} is dispatching VOICE_UPDATE.")

    @classmethod
    async def connect_to_channel(cls, channel: VocalGuildChannel, /, **kwargs) -> Player:
        player = await channel.connect(cls=cls)
        player.add_to_namespace(kwargs)

        return player

    async def connect(
        self, *, timeout: float = 10.0, reconnect: bool, self_deaf: bool = True, self_mute: bool = False
    ) -> None:
        """
        Do not use this method directly on the player. See: :meth:`disnake.VoiceChannel.connect` for more details.
        Pass the :class:`Player` to ``cls=`` in :meth:`disnake.VoiceChannel.connect`.


        Raises
        ------
        ChannelTimeoutException
            Connecting to the voice channel timed out.
        InvalidChannelStateException
            You tried to connect this player without an appropriate voice channel.
        """
        if self.channel is MISSING:
            msg: str = 'Please use "disnake.VoiceChannel.connect(cls=...)" and pass this Player to cls.'
            raise InvalidChannelStateException(f"Player tried to connect without a valid channel: {msg}")

        if not self._guild:
            self._guild = self.channel.guild

        self.node.players[self._guild.id] = self

        assert self.guild is not None
        await self.guild.change_voice_state(channel=self.channel, self_mute=self_mute, self_deaf=self_deaf)

        try:
            async with async_timeout.timeout(timeout):
                await self._connection_event.wait()
        except (asyncio.TimeoutError, asyncio.CancelledError):
            msg = f"Unable to connect to {self.channel} as it exceeded the timeout of {timeout} seconds."
            raise ChannelTimeoutException(msg)

    async def move_to(
        self,
        channel: VocalGuildChannel | None,
        *,
        timeout: float = 10.0,
        self_deaf: bool | None = None,
        self_mute: bool | None = None,
    ) -> None:
        """Method to move the player to another channel.

        Parameters
        ----------
        channel: :class:`disnake.VoiceChannel` | :class:`disnake.StageChannel`
            The new channel to move to.
        timeout: float
            The timeout in ``seconds`` before raising. Defaults to 10.0.
        self_deaf: bool | None
            Whether to deafen when moving. Defaults to ``None`` which keeps the current setting or ``False``
            if they can not be determined.
        self_mute: bool | None
            Whether to self mute when moving. Defaults to ``None`` which keeps the current setting or ``False``
            if they can not be determined.

        Raises
        ------
        ChannelTimeoutException
            Connecting to the voice channel timed out.
        InvalidChannelStateException
            You tried to connect this player without an appropriate guild.
        """
        if not self.guild:
            raise InvalidChannelStateException("Player tried to move without a valid guild.")

        self._connection_event.clear()
        voice: disnake.VoiceState | None = self.guild.me.voice

        if self_deaf is None and voice:
            self_deaf = voice.self_deaf

        if self_mute is None and voice:
            self_mute = voice.self_mute

        self_deaf = bool(self_deaf)
        self_mute = bool(self_mute)

        await self.guild.change_voice_state(channel=channel, self_mute=self_mute, self_deaf=self_deaf)

        if channel is None:
            return

        try:
            async with async_timeout.timeout(timeout):
                await self._connection_event.wait()
        except (asyncio.TimeoutError, asyncio.CancelledError):
            msg = f"Unable to connect to {channel} as it exceeded the timeout of {timeout} seconds."
            raise ChannelTimeoutException(msg)

    async def play(
        self,
        track: Playable,
        *,
        replace: bool = True,
        start: int = 0,
        end: int | None = None,
        volume: int | None = None,
        paused: bool | None = None,
        add_history: bool = True,
        filters: Filters | None = None,
        clear_auto: bool = False
    ) -> Playable:
        """Play the provided :class:`~Playable`.

        Parameters
        ----------
        track: :class:`~Playable`
            The track to being playing.
        replace: bool
            Whether this track should replace the currently playing track, if there is one. Defaults to ``True``.
        start: int
            The position to start playing the track at in milliseconds.
            Defaults to ``0`` which will start the track from the beginning.
        end: Optional[int]
            The position to end the track at in milliseconds.
            Defaults to ``None`` which means this track will play until the very end.
        volume: Optional[int]
            Sets the volume of the player. Must be between ``0`` and ``1000``.
            Defaults to ``None`` which will not change the current volume.
            See Also: :meth:`set_volume`
        paused: bool | None
            Whether the player should be paused, resumed or retain current status when playing this track.
            Setting this parameter to ``True`` will pause the player. Setting this parameter to ``False`` will
            resume the player if it is currently paused. Setting this parameter to ``None`` will not change the status
            of the player. Defaults to ``None``.
        add_history: Optional[bool]
            If this argument is set to ``True``, the :class:`~Player` will add this track into the
            :class:`Queue` history, if loading the track was successful. If ``False`` this track will not be
            added to your history. This does not directly affect the ``AutoPlay Queue`` but will alter how ``AutoPlay``
            recommends songs in the future. Defaults to ``True``.
        filters: Optional[:class:`~Filters`]
            An Optional[:class:`~Filters`] to apply when playing this track. Defaults to ``None``.
            If this is ``None`` the currently set filters on the player will be applied.
        clear_auto: Optional[bool]
            If user wants to clear the ``AutoPlay Queue`` after the track has been played. Defaults to `False`

        Returns
        -------
        :class:`~Playable`
            The track that began playing.
        """
        assert self.guild is not None

        if volume and self._volume != volume:
            self._volume = volume

        if replace or not self._current:
            self._current = track
            self._original = track

        old_previous = self._previous
        self._previous = self._current
        self.queue._loaded = track

        if paused is not None:
            pause = paused
        else:
            pause = self._paused

        if filters:
            self._filters = filters

        if clear_auto:
            self.auto_queue.clear()

        request: RequestPayload = {
            "track": {"encoded": track.encoded, "userData": dict(track.extras)},
            "volume": int(self._volume),
            "position": start,
            "endTime": end,
            "paused": pause,
            "filters": self._filters(),
        }

        try:
            await self.node.update(self.guild.id, data=request, replace=replace)
        except LavalinkException as e:
            self._current = None
            self._original = None
            self._previous = old_previous
            self.queue._loaded = old_previous
            self._volume = self._volume
            raise e

        self._paused = pause

        if add_history:
            assert self.queue.history is not None
            self.queue.history.put(track)

        self.dispatch_message_update()
        self.paused_cancel()
        return track

    def paused_cancel(self) -> None:
        if self._paused_task:
            try:
                self._paused_task.cancel()
            except Exception as e:
                _ = e
                pass

        self._paused_task = None

    def paused_start(self) -> None:
        if self._paused_wait > 0:
            self._paused_task = asyncio.create_task(self._runner(self._paused_wait))
            self._paused_task.add_done_callback(self._inactivity_task_callback)

    async def pause(self, value: bool, /) -> None:
        """Set the paused or resume state of the player.

        Parameters
        ----------
        value: bool
            A bool indicating whether the player should be paused or resumed. True indicates that the player should be
            ``paused``. False will resume the player if it is currently paused.
        """
        assert self.guild is not None

        request: RequestPayload = {"paused": value}
        await self.node.update(self.guild.id, data=request)

        if value:
            self.paused_start()
        else:
            self.paused_cancel()

        self._paused = value

    async def resume(self, value: bool, /) -> None:
        """Set the resume state of the player.

        Parameters
        ----------
        value: bool
            A bool indicating whether the player should be resumed. True indicates that the player should be
            ``resumed``. False will pause the player if it is currently playing.
        """
        return await self.pause(not value)

    async def seek(self, position: int = 0, /) -> None:
        """Seek to the provided position in the currently playing track, in milliseconds.

        Parameters
        ----------
        position: int
            The position to seek to in milliseconds. To restart the song from the beginning,
            you can disregard this parameter or set position to 0.
        """
        assert self.guild is not None

        if not self._current:
            return

        request: RequestPayload = {"position": position}
        await self.node.update(self.guild.id, data=request)
        self.dispatch_message_update()

    async def set_filters(self, filters: Filters | None = None, /, *, seek: bool = True) -> None:
        """Set the :class:`Filters` on the player.

        Parameters
        ----------
        filters: Optional[:class:`~Filters`]
            The filters to set on the player. Could be ``None`` to reset the currently applied filters.
            Defaults to ``None``.
        seek: bool
            Whether to seek immediately when applying these filters. Seeking uses more resources, but applies the
            filters immediately. Defaults to ``False``.

        This method now accepts a positional-only argument of filters, which now defaults to None. Filters
        were redesigned in this version, see: :class:`Filters`.
        """
        assert self.guild is not None

        if filters is None:
            filters = Filters()

        request: RequestPayload = {"filters": filters()}
        await self.node.update(self.guild.id, data=request)
        self._filters = filters

        if self.playing and seek:
            await self.seek(self.position)

        self.dispatch_message_update()

    async def set_volume(self, value: int = 100, /) -> None:
        """Set the :class:`Player` volume, as a percentage, between 0 and 1000.

        By default, every player is set to 100 on creation. If a value outside 0 to 1000 is provided it will be
        clamped.

        Parameters
        ----------
        value: int
            A volume value between 0 and 1000. To reset the player to 100, you can disregard this parameter.

        The ``value`` parameter is now positional-only, and has a default of 100.
        """
        assert self.guild is not None
        vol: int = max(min(value, 1000), 0)

        request: RequestPayload = {"volume": vol}
        await self.node.update(self.guild.id, data=request)

        self._volume = vol
        self.dispatch_message_update()

    async def disconnect(self, **kwargs: Any) -> None:
        """Disconnect the player from the current voice channel and remove it from the :class:`~Node`.

        This method will cause any playing track to stop and potentially trigger the following events:

            - ``on_mystic_track_end``
            - ``on_mystic_websocket_closed``


        .. warning::

            Please do not re-use a :class:`Player` instance that has been disconnected, unwanted side effects are
            possible.
        """
        assert self.guild

        try:
            await self.destroy()
        except Exception as e:
            _ = e
        finally:
            await self.guild.change_voice_state(channel=None)

    async def skip(self, *, force: bool = True) -> Playable | None:
        """Stop playing the currently playing track.

        Parameters
        ----------
        force: bool
            Whether the track should skip looping, if :class:`Queue` has been set to loop.
            Defaults to ``True``.

        Returns
        -------
        :class:`~Playable` | None
            The currently playing track that was skipped, or ``None`` if no track was playing.
        """
        assert self.guild is not None
        old: Playable | None = self._current

        if force:
            self.queue._loaded = None

        request: RequestPayload = {"track": {"encoded": None}}
        await self.node.update(self.guild.id, data=request, replace=True)
        self.dispatch_message_update()

        return old

    def _invalidate(self) -> None:
        self._connected = False
        self._connection_event.clear()
        self.inactivity_cancel()

        try:
            self.cleanup()
        except (AttributeError, KeyError):
            pass

    async def destroy(self) -> None:
        assert self.guild

        self._invalidate()
        player: Player | None = self.node.players.pop(self.guild.id, None)
        self.node.client.dispatch("mystic_player_destroyed", player)

        if player:
            try:
                await self.node.destroy(self.guild.id)
            except LavalinkException:
                pass
